/*
 * This work is based on a linux driver and the hardware specs 
 * available here: http://embuildsw.org/iBurstModemDriverGuide.pdf
 * 
 * Big thanks go to:
 * 1. Leonid Gojayev for providing the modem for development and testing
 * 2. Alexander V. Chernikov aka bird_of_luck for his invaluable help and useful advices
 * 3. FreeBSD mailing lists
 *
 *
 * This driver is provided under the term of the GNU GPLv2 license.
 *
 * (c) Tofig Suleymanov  20 February, 2008
 */


#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_media.h>
#include <net/if_dl.h>
#include <net/if_types.h>

#include <sys/sysctl.h>
#include <net/bpf.h>


#include <sys/bus.h>
#include <machine/bus.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbdivar.h>
#include "usbdevs.h"
#include <dev/usb/usb_ethersubr.h>



#include <../../dev/ib/if_ib.h>

#define DPRINTF(x) printf x


/*
 * USB configuration frame.
 * the rightmost byte can be 0x02 which is networking-only mode, and 0x06 which is networking + status messages mode 
 */

static unsigned char ib_usb_conf[] = {
        0x00, 0x08, 0x00, 0xf7, IB_PROTO, PROTO_USB_CONF, 0x00, 0x06
};



struct ib_softc {
 	struct ifnet		*ib_ifp;
 	device_t 		ib_dev;
	struct ifmedia          ib_ifmedia;
	usbd_device_handle 	ib_udev;
	usbd_interface_handle 	ib_iface;
        usbd_pipe_handle        ib_bulkin_ep, ib_bulkout_ep; 
        struct ue_cdata         ib_cdata;
        struct callout_handle   ib_stat_ch;
	struct usb_qdat		ib_qdat;
        struct mtx              ib_mtx;
	u_char			modemhw_addr[ETHER_ADDR_LEN];
	u_char			hosthw_addr[ETHER_ADDR_LEN];
	int                     ib_bulkin_no, ib_bulkout_no;
        int                     ib_unit;
        int                     ib_if_flags;
	int 			ib_control;
	int			ib_if_status;
	int			ib_rxeof_errors;
	int			ib_rx_packets;
	int			ib_tx_packets;
	unsigned int		ib_signal_strength;
	unsigned char 		ib_ctl_buf[32];
        char 			ib_dying;
	
};

static device_probe_t ib_match;
static device_attach_t ib_attach;
static device_detach_t ib_detach;
static device_shutdown_t ib_shutdown;
static int ib_ioctl(struct ifnet *, u_long, caddr_t);

static void ib_start(struct ifnet *);
static void ib_init(void *);
static void ib_watchdog(struct ifnet *);
static void ib_rxstart(struct ifnet *);

static void ib_ifmedia_sts(struct ifnet *ifp, struct ifmediareq *ifmr);
static void ib_send_config_frame(struct ib_softc *sc);
static void ib_rxeof(usbd_xfer_handle, usbd_private_handle, usbd_status);
static void ib_txeof(usbd_xfer_handle, usbd_private_handle, usbd_status);
static void ib_stop(struct ib_softc *);

static int ib_ifmedia_upd(struct ifnet *ifp);
static int ib_encap(struct ib_softc *, struct mbuf *, int);

usbd_status ib_getmac(struct ib_softc *, int , unsigned char *);

//void ib_reset(struct ib_softc *);


static device_method_t ib_methods[] = {
        /* Device interface */
        DEVMETHOD(device_probe,         ib_match),
        DEVMETHOD(device_attach,        ib_attach),
        DEVMETHOD(device_detach,        ib_detach),
        DEVMETHOD(device_shutdown,      ib_shutdown),
        { 0, 0 }
};

static driver_t ib_driver = {
        "ib",
        ib_methods,
        sizeof(struct ib_softc)
};

static devclass_t ib_devclass;

DRIVER_MODULE(ib, uhub, ib_driver, ib_devclass, usbd_driver_load, 0);
MODULE_DEPEND(ib, usb, 1, 1, 1);
MODULE_DEPEND(ib, ether, 1, 1, 1);


static int ib_match(device_t self)
{
	struct usb_attach_arg *uaa = device_get_ivars(self);
	usb_device_descriptor_t *dd;

	DPRINTF(("ib_match\n"));
	if (!uaa->iface)
		return UMATCH_NONE;

	dd = usbd_get_device_descriptor(uaa->device);

	if (dd &&
	    ((UGETW(dd->idVendor) == 0x0482 &&
	    UGETW(dd->idProduct) == 0x0204)))
		return UMATCH_VENDOR_PRODUCT;
	else
		return UMATCH_NONE;
}

static int ib_attach(device_t self)
{

	struct ib_softc *sc = device_get_softc(self);
	struct usb_attach_arg *uaa = device_get_ivars(self);
	char devinfo[1024];
	usbd_status status;
	usbd_interface_handle   iface;
	usb_interface_descriptor_t *id;
        usb_device_descriptor_t *dd;
        usb_endpoint_descriptor_t *ed;
	struct ifnet *ifp;
	int i;

	DPRINTF(("ib_attach: sc=%p\n", sc));

	bzero(sc, sizeof(struct ib_softc));

 	if ((!uaa->device) || (!uaa->iface)) {
		DPRINTF(("no device or iface"));
		return ENXIO;
	}

	usbd_devinfo(uaa->device, 0, devinfo);
        device_set_desc_copy(self, devinfo);
        printf("%s: %s\n", device_get_nameunit(self), devinfo);

	sc->ib_dev = self;
        sc->ib_udev = uaa->device;
        sc->ib_unit = device_get_unit(self);

	if (usbd_set_config_no(sc->ib_udev, 1, 0)) {
                printf("ib%d: setting configuration failed\n",
                    sc->ib_unit);
                return ENXIO;
        }


	if(usbd_device2interface_handle(uaa->device, 0, &iface)) {
                printf("ib%d: getting interface handle failed\n",
                    sc->ib_unit);
                return ENXIO;
        }

	sc->ib_iface = iface;

/*
	if (usbd_set_interface(sc->ib_iface, 0)) {
        	printf("%s: setting  interface failed\n",
               	device_get_nameunit(sc->ib_dev));
                return ENXIO;
	}
*/

	if(sc->ib_iface->idesc) {
		DPRINTF(("sc->ib_iface->idesc->bNumEndpoints: %d\n", sc->ib_iface->idesc->bNumEndpoints));
	}

        dd = usbd_get_device_descriptor(sc->ib_udev);
        switch (UGETW(dd->bcdDevice)) {
        case UT_DEVICE_UT_02:
                sc->ib_control = CONTROL_USB_GET_MAC_ADDR_UT_02;
                DPRINTF(("if_ib: UT_02 Detected\n"));
                break;
        case UT_DEVICE_UT_04:
                sc->ib_control = CONTROL_USB_GET_MAC_ADDR_UT_04;
                DPRINTF(("if_ib: UT_04 Detected\n"));
                break;
        default:
                DPRINTF(("if_ib: Unknown UT bcdDevice %04x\n", UGETW(dd->bcdDevice)));
                return -ENODEV;
        }


        id = usbd_get_interface_descriptor(sc->ib_iface);

        /* Get endpoints. */
        for (i = 0; i < id->bNumEndpoints; i++) {
                ed = usbd_interface2endpoint_descriptor(iface, i);
                if (!ed) {
                        printf("ib%d: couldn't get ed %d\n",
                            sc->ib_unit, i);
                	return ENXIO;
                }
                if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
                    UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
                        sc->ib_bulkin_no = ed->bEndpointAddress;
			DPRINTF(("RX ed: %x\n", sc->ib_bulkin_no));
                } else if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_OUT &&
                           UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
                        sc->ib_bulkout_no = ed->bEndpointAddress;
			DPRINTF(("TX ed: %x\n", sc->ib_bulkout_no));
                }
        }


        mtx_init(&sc->ib_mtx, device_get_nameunit(self), MTX_NETWORK_LOCK, MTX_DEF | MTX_RECURSE);
	ifmedia_init(&sc->ib_ifmedia, 0, ib_ifmedia_upd, ib_ifmedia_sts);

	// XXX should reset the device here. It seems to be a standard SET_CONFIG (0x09) request,

	status = ib_getmac(sc, CONTROL_USB_GET_MAC_ADDR_UT_02, sc->ib_ctl_buf);
	if(status != USBD_NORMAL_COMPLETION) {
		DPRINTF(("ib_getmac: invalid read %02x %02x\n",
				 sc->ib_ctl_buf[0], CONTROL_USB_GET_MAC_ADDR_UT_02));
                mtx_destroy(&sc->ib_mtx);
                return ENXIO;
	}	

	if (sc->ib_ctl_buf[1] != sc->ib_control) {
	DPRINTF(("ib: mismatch command %02x %02x\n",
			sc->ib_ctl_buf[1], sc->ib_control));
			mtx_destroy(&sc->ib_mtx);
                	return ENXIO;
        }

	ifp = sc->ib_ifp = if_alloc(IFT_ETHER);
        if (ifp == NULL) {
                printf("ib%d: can not if_alloc()\n", sc->ib_unit);
                mtx_destroy(&sc->ib_mtx);
                return ENXIO;
        }

        ifp->if_softc = sc;
        if_initname(ifp, "ib", sc->ib_unit);
        ifp->if_mtu = ETHERMTU;
        ifp->if_flags = IFF_BROADCAST |  IFF_SIMPLEX | IFF_NEEDSGIANT;
        ifp->if_ioctl = ib_ioctl;
        ifp->if_start = ib_start;
        ifp->if_watchdog = ib_watchdog;
        ifp->if_init = ib_init;
        ifp->if_baudrate = 1000000; // According to specs it is 1024k
        ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;

        sc->ib_qdat.ifp = ifp;
        sc->ib_qdat.if_rxstart = ib_rxstart;
	
        /* No IFM type for 11Mbps USB, so go with 10baseT */
        ifmedia_add(&sc->ib_ifmedia, IFM_ETHER | IFM_10_T, 0, 0);
        ifmedia_set(&sc->ib_ifmedia, IFM_ETHER | IFM_10_T);


	// initialize modem hw addr
        memcpy(sc->modemhw_addr, sc->ib_ctl_buf + 2, ETHER_ADDR_LEN);

	sc->ib_ctl_buf[2 + ETHER_ADDR_LEN - 1] ^= 1; // flip the LSB of last byte to get host MAC addr
        memcpy(sc->hosthw_addr, sc->ib_ctl_buf + 2, ETHER_ADDR_LEN);

        sc->ib_dying = 0;
	sc->ib_rx_packets = 0;
	sc->ib_tx_packets = 0;
        sc->ib_rxeof_errors = 0;
	sc->ib_signal_strength = 0; // can be from 0 to 100

	SYSCTL_ADD_UINT(device_get_sysctl_ctx(sc->ib_dev),
            SYSCTL_CHILDREN(device_get_sysctl_tree(sc->ib_dev)),
            OID_AUTO, "signal_level", CTLFLAG_RD, &sc->ib_signal_strength, 0,
            "Signal strength [0 - 100] ");


        ether_ifattach(ifp, sc->hosthw_addr);
        usb_register_netisr();
	return 0;
}


static int ib_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	struct ib_softc        *sc = ifp->if_softc;
	struct ifreq            *ifr = (struct ifreq *)data;

	int error = 0;

	//DPRINTF("ib_ioctl reached\n");

	switch(cmd) {

        case SIOCSIFFLAGS:
                if (ifp->if_flags & IFF_UP) {
                        if (!(ifp->if_drv_flags & IFF_DRV_RUNNING))
                                ib_init(sc);
                } else {
                        if (ifp->if_drv_flags & IFF_DRV_RUNNING)
                                ib_stop(sc);
                }

                error = 0;
                break;

	case SIOCSIFMEDIA:
	case SIOCGIFMEDIA:
                error = ifmedia_ioctl(ifp, ifr, &sc->ib_ifmedia, cmd);
                break;
        default:
                error = ether_ioctl(ifp, cmd, data);
                break;
        }

	return error;
}

usbd_status ib_getmac(struct ib_softc *sc, int ut_type, unsigned char *buffer)
{
        usb_device_request_t req;
        req.bmRequestType = UT_READ_VENDOR_DEVICE;
        req.bRequest = ut_type;
	USETW(req.wValue, 0); 
        USETW(req.wIndex, 0);
        USETW(req.wLength, 8);
        return usbd_do_request(sc->ib_udev, &req, buffer);
}


static int 
ib_shutdown(device_t dev)
{
//	DPRINTF("ib_shutdown entered\n");
        struct ib_softc *sc;

        sc = device_get_softc(dev);
        ib_stop(sc);

        return (0);
}

void 
ib_start(struct ifnet *ifp)
{

        struct ib_softc        *sc;
        struct mbuf             *m_head = NULL;

	//DPRINTF("ib_start entered\n");

        sc = ifp->if_softc;

	if (! sc->ib_if_status & STATUS_READY) {
                return;
        }

        if (ifp->if_drv_flags & IFF_DRV_OACTIVE) {
                return;
        }

        IFQ_DRV_DEQUEUE(&ifp->if_snd, m_head);
        if (m_head == NULL) {
                return;
        }

        if (ib_encap(sc, m_head, 0)) {
                IFQ_DRV_PREPEND(&ifp->if_snd, m_head);
                ifp->if_drv_flags |= IFF_DRV_OACTIVE;
                return;
        }

	BPF_MTAP(ifp, m_head);

	ifp->if_drv_flags |= IFF_DRV_OACTIVE;
	
	ifp->if_timer = 5;
	return;

	
}

void ib_init(void *ptr) 
{
	//DPRINTF("ib_init entered\n");

        struct ib_softc       	*sc = ptr;
        struct ifnet            *ifp = sc->ib_ifp;
        struct ue_chain 	*c;
        usbd_status             err;
        int                     i;

        if (ifp->if_drv_flags & IFF_DRV_RUNNING)
                return;

        //ib_reset(sc);


        if (usb_ether_tx_list_init(sc, &sc->ib_cdata,
            sc->ib_udev) == ENOBUFS) {
                printf("%s: tx list init failed\n", device_get_nameunit(sc->ib_dev));
		mtx_destroy(&sc->ib_mtx);
                return;
        }

        if (usb_ether_rx_list_init(sc, &sc->ib_cdata,
            sc->ib_udev) == ENOBUFS) {
                printf("%s: rx list init failed\n", device_get_nameunit(sc->ib_dev));
		mtx_destroy(&sc->ib_mtx);
                return;
        }


        err = usbd_open_pipe(sc->ib_iface, sc->ib_bulkin_no,
            USBD_EXCLUSIVE_USE, &sc->ib_bulkin_ep);
        if (err) {
                printf("%s: open rx pipe failed: %s\n", device_get_nameunit(sc->ib_dev),
                    usbd_errstr(err));
		mtx_destroy(&sc->ib_mtx);
                return;
        }

        err = usbd_open_pipe(sc->ib_iface, sc->ib_bulkout_no,
            USBD_EXCLUSIVE_USE, &sc->ib_bulkout_ep);
        if (err) {
                printf("%s: open tx pipe failed: %s\n", device_get_nameunit(sc->ib_dev),
                    usbd_errstr(err));
		mtx_destroy(&sc->ib_mtx);
                return;
        }
	


        for (i = 0; i < UE_RX_LIST_CNT; i++) {
                c = &sc->ib_cdata.ue_rx_chain[i];
                usbd_setup_xfer(c->ue_xfer, sc->ib_bulkin_ep, c,
                    mtod(c->ue_mbuf, char *), UE_BUFSZ, USBD_SHORT_XFER_OK,
                    USBD_NO_TIMEOUT, ib_rxeof);
                usbd_transfer(c->ue_xfer);
        }



	ifp->if_drv_flags |=  IFF_DRV_RUNNING;
	ifp->if_drv_flags &=  ~IFF_DRV_OACTIVE;


	ib_send_config_frame(sc);

	/* We do not query for carrier status. We just dumbly
	 * activate it right after sending the config frame
	 */
	sc->ib_if_status |= STATUS_READY;

}


void ib_watchdog(struct ifnet *ifp) 
{
	/* stub */
	DPRINTF(("ib_watchdog entered\n"));
}
	
void
ib_rxstart(struct ifnet *ifp)
{
        struct ib_softc        *sc;
        struct ue_chain *c;
	//DPRINTF(("ib_rxstart entered\n")); 

        sc = ifp->if_softc;

        c = &sc->ib_cdata.ue_rx_chain[sc->ib_cdata.ue_rx_prod];
 
        c->ue_mbuf = usb_ether_newbuf();
        if (c->ue_mbuf == NULL) {
                printf("%s: no memory for rx list "
                    "-- packet dropped!\n", device_get_nameunit(sc->ib_dev));
                ifp->if_ierrors++;
                return;
        }

        /* Setup new transfer. */
        usbd_setup_xfer(c->ue_xfer, sc->ib_bulkin_ep,
            c, mtod(c->ue_mbuf, char *), UE_BUFSZ, USBD_SHORT_XFER_OK,
            USBD_NO_TIMEOUT, ib_rxeof);
        usbd_transfer(c->ue_xfer);

       return;
}

static int ib_detach(device_t dev)
{
	struct ib_softc        *sc;
        struct ifnet            *ifp; 

	sc = device_get_softc(dev);
        ifp = sc->ib_ifp;


	sc->ib_dying = 1;
	ether_ifdetach(ifp);
	if_free(ifp);
	ifmedia_removeall(&sc->ib_ifmedia);

	if (sc->ib_bulkout_ep != NULL)
	usbd_abort_pipe(sc->ib_bulkout_ep);
	if (sc->ib_bulkin_ep != NULL)
	usbd_abort_pipe(sc->ib_bulkout_ep);

        mtx_destroy(&sc->ib_mtx);
 
	DPRINTF(("ib_detach: sc=%p\n", sc));
	return (0);
}

static int
ib_ifmedia_upd(struct ifnet *ifp)
{

	/* this is a stub yet */
	DPRINTF(("ib_ifmedia_upd entered "));

        return 0;
}

void
ib_ifmedia_sts(struct ifnet * const ifp, struct ifmediareq *req)
{
	struct ib_softc *sc = ifp->if_softc;


	if(sc->ib_if_status & STATUS_READY) 
  			req->ifm_status = IFM_ACTIVE;
		else 
			req->ifm_status &= ~IFM_ACTIVE;


	req->ifm_status |= IFM_AVALID;
	req->ifm_active = IFM_ETHER | IFM_10_T;

}


static void 
ib_rxeof(usbd_xfer_handle xfer, usbd_private_handle priv, usbd_status status)
{
	struct ue_chain 	*c = priv;
        struct ib_softc       	*sc = c->ue_sc;
        struct ifnet            *ifp;
        struct mbuf             *m;
	struct ib_net_radio_t 	*radio; 
	struct ib_net_ether_t	*packet;
        int			total_len = 0;


	//DPRINTF("rxeof entered\n");
        ifp = sc->ib_ifp;

        if (sc->ib_dying || !(ifp->if_drv_flags & IFF_DRV_RUNNING)) {
                return;
        }

        if (status != USBD_NORMAL_COMPLETION) {
                if (status == USBD_NOT_STARTED || status == USBD_CANCELLED) {
                        return;
                }
                if (sc->ib_rxeof_errors == 0)
                        printf("%s: usb error on rx: %s\n",
                            device_get_nameunit(sc->ib_dev), usbd_errstr(status));
                if (status == USBD_STALLED)
                        usbd_clear_endpoint_stall_async(sc->ib_bulkin_ep);
                DELAY(sc->ib_rxeof_errors * 10000);
                sc->ib_rxeof_errors++;
                goto done;
        }


        usbd_get_xfer_status(xfer, NULL, NULL, &total_len, NULL);
	radio = mtod(c->ue_mbuf, struct ib_net_radio_t *);


	/* We drop packets with FLAG_EXTENSION bit */
        if (radio->word[0] & FLAG_EXTENSION) {
		DPRINTF(("(dropped FLAG_EXTENSION packet)\n"));
		goto done;
	}
	/* We drop packets without 1's complement */
        if (radio->word[1] ^ radio->check ^ 0xff) { 
		DPRINTF(("(dropped packet.Incorrect complement)\n"));
		goto done;
	}


	/* Pass a network packet to the upper layer */
        if (radio->proto[0] != IB_PROTO) { 
//		DPRINTF("(Network packet caught)\n");
		m = usb_ether_newbuf();
		if (m == NULL) {
			printf("%s: no memory for rx mbuf chain "
			    "-- packet dropped!\n", 
				device_get_nameunit(sc->ib_dev));
				ifp->if_ierrors++;
				goto done;
        	}
		m->m_pkthdr.rcvif = (struct ifnet *)&sc->ib_qdat;
		m->m_pkthdr.len = m->m_len = total_len - IB_NET_RADIO_HEAD + sizeof(struct ib_net_ether_t); 

		packet = mtod(m, struct ib_net_ether_t *);

		memcpy(packet->source, sc->modemhw_addr, ETHER_ADDR_LEN);

		if(radio->word[0] & 0x08) // broadcast ?
			memset(packet->dest, 0xff, ETHER_ADDR_LEN);
		else
			memcpy(packet->dest,
				 sc->hosthw_addr, ETHER_ADDR_LEN);

		packet->proto[0] = radio->proto[0];
		packet->proto[1] = radio->proto[1];


		memcpy(packet->payload, radio->payload, total_len - IB_NET_RADIO_HEAD); 


		sc->ib_rx_packets++;
		ifp->if_ipackets++;

		/* Put the packet on the special USB input queue. */
		usb_ether_input(m);
		return;
	}
        else if (radio->proto[1] == IB_PROTO_CONTROL) {
		/* this is a status1 packet which contains the signal strength */
		if(radio->payload[7] == STATUS1) { 
			/* update signal strength sysctl value */
			sc->ib_signal_strength = radio->payload[8];
		}
		goto done;
	}

done:
        /* Setup new transfer. */
        usbd_setup_xfer(c->ue_xfer, sc->ib_bulkin_ep, c,
            mtod(c->ue_mbuf, char *),
            UE_BUFSZ, USBD_SHORT_XFER_OK, USBD_NO_TIMEOUT,
            ib_rxeof);
        usbd_transfer(c->ue_xfer);

        return;
}


void 
ib_stop(struct ib_softc *sc)
{
	//DPRINTF("ib_stop entered\n");
	usbd_status      err;
        struct ifnet    *ifp;

        //cdce_reset(sc);

        ifp = sc->ib_ifp;

        //ifp->if_timer = 0;

        if (sc->ib_bulkin_ep != NULL) {
                err = usbd_abort_pipe(sc->ib_bulkin_ep);
                if (err)
                        printf("%s: abort rx pipe failed: %s\n",
                            device_get_nameunit(sc->ib_dev), usbd_errstr(err));
                err = usbd_close_pipe(sc->ib_bulkin_ep);
                if (err)
                        printf("%s: close rx pipe failed: %s\n",
                            device_get_nameunit(sc->ib_dev), usbd_errstr(err));
                sc->ib_bulkin_ep = NULL;
        }
        if (sc->ib_bulkout_ep != NULL) {
                err = usbd_abort_pipe(sc->ib_bulkout_ep);
                if (err)
                        printf("%s: abort tx pipe failed: %s\n",
                            device_get_nameunit(sc->ib_dev), usbd_errstr(err));
                err = usbd_close_pipe(sc->ib_bulkout_ep);
                if (err)
                        printf("%s: close tx pipe failed: %s\n",
                            device_get_nameunit(sc->ib_dev), usbd_errstr(err));
                sc->ib_bulkout_ep = NULL;
        }

        usb_ether_rx_list_free(&sc->ib_cdata);
        usb_ether_tx_list_free(&sc->ib_cdata);

        ifp->if_drv_flags &= ~(IFF_DRV_RUNNING | IFF_DRV_OACTIVE);

	sc->ib_if_status &= ~STATUS_READY;

	return;
}


void
ib_send_config_frame(struct ib_softc *sc)
{
	usbd_status err;
	struct ue_chain         *c;
	struct ib_net_radio_t	*radio;

	DPRINTF(("ib_send_config_frame entered\n"));

	c = &sc->ib_cdata.ue_tx_chain[0];

	radio = (struct ib_net_radio_t*) &ib_usb_conf;

	radio->payload[0] = sc->ib_tx_packets & 0xff;
	radio->payload[0] |= STATUS_READY;
	usbd_setup_xfer(c->ue_xfer, sc->ib_bulkout_ep, c,
            &ib_usb_conf,
            sizeof(ib_usb_conf), USBD_FORCE_SHORT_XFER, 10000,
            ib_txeof);
        err = usbd_transfer(c->ue_xfer);

	DELAY(10000);
        if (err != USBD_IN_PROGRESS) 
		DPRINTF(("ERROR TX OF CONFIG BYTES\n"));
        sc->ib_cdata.ue_tx_cnt++;

	sc->ib_tx_packets++;
}	


void 
ib_txeof(usbd_xfer_handle xfer , usbd_private_handle priv, usbd_status status)
{

        struct ib_softc         *sc;
        struct ue_chain *c;
        struct ifnet            *ifp;
        usbd_status             err;
	int total_len = 0 ;

	//DPRINTF("txeof entered \n");

        c = priv;
        sc = c->ue_sc;
        ifp = sc->ib_ifp;

        ifp->if_timer = 0;


        if (status != USBD_NORMAL_COMPLETION) {
                if (status == USBD_NOT_STARTED || status == USBD_CANCELLED) {
                        return;
                }
                printf("ib%d: usb error on tx: %s\n", sc->ib_unit,
                    usbd_errstr(status));
                if (status == USBD_STALLED)
                        usbd_clear_endpoint_stall(sc->ib_bulkout_ep);
                return;
        }


        ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
        usbd_get_xfer_status(c->ue_xfer, NULL, NULL, &total_len, &err);

	if (c->ue_mbuf != NULL) {
                c->ue_mbuf->m_pkthdr.rcvif = ifp;
                usb_tx_done(c->ue_mbuf);
                c->ue_mbuf = NULL;
        }


//	DPRINTF("total encap length: %d\n", total_len);
        if (err)
                ifp->if_oerrors++;
        else {
                ifp->if_opackets++;
		sc->ib_tx_packets++;
	}
        return;
}


static int 
ib_encap(struct ib_softc *sc, struct mbuf *m, int idx) 
{
	struct ue_chain 	*c;
        usbd_status             err;
	struct ether_header 	*hdr;
	struct ib_net_radio_t	*radio;
	int total_len;
	//DPRINTF("ib_encap entered\n");

        c = &sc->ib_cdata.ue_tx_chain[idx];

	hdr = mtod(m, struct ether_header *);

        /*
         * Copy the mbuf data into a contiguous buffer, stripping
	 * sizeof(ib_net_radio_t) - sizeof(ether_header) 
         * bytes from the beginning to hold the radio protocol
         */
        m_copydata(m, 0, m->m_pkthdr.len,
			 c->ue_buf + 
		sizeof(struct ib_net_radio_t) - sizeof(struct ether_header));
        c->ue_mbuf = m;

	total_len = m->m_pkthdr.len + sizeof(struct ib_net_radio_t) - sizeof(struct ether_header);

	radio = (struct ib_net_radio_t *) c->ue_buf;

	radio->word[0] = (total_len) >> 8;
	if (hdr->ether_dhost[0] & 1) // broadcast ?
                radio->word[0] |= 8;
	radio->word[1] = (total_len) & 0xff;

	/* construct 1's complement */
        radio->check = radio->word[1] ^ 0xff;

	/* construct radio->proto. It should be big-endian (network byte order) */
        radio->proto[0] = hdr->ether_type;
        radio->proto[1] = hdr->ether_type >> 8;

	//DPRINTF("SENT PACKETS %d\n" , total_len);
        usbd_setup_xfer(c->ue_xfer, sc->ib_bulkout_ep,
            c, c->ue_buf, total_len, USBD_FORCE_SHORT_XFER,
            10000, ib_txeof);

        /* Transmit */
        err = usbd_transfer(c->ue_xfer);
        if (err != USBD_IN_PROGRESS) {
                ib_stop(sc);
                return(EIO);
        }

        sc->ib_cdata.ue_tx_cnt++;
        return(0);
}

